/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

/*
 * Copyright: Philipp Kolmann (c) 2006
 *            philipp@kolmann.at
 * 
 * Purpose: Enable Vol+/- Buttons on USB Headset
 *
 * inspired by: http://www.fuw.edu.pl/~pliszka/linux-USB/#labtec
 *
 */

/* Changelog:
 *
 * 0.0.1: 2005-01-05
 *     Initial Release
 *
 * 0.0.2: 2005-01-08
 *     Add startup options and make more error aware
 *
*/

#define VERSION "0.0.2"

#include <alsa/asoundlib.h>
#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <ctype.h>

#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <asm/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdint.h>

#include <linux/input.h>



snd_ctl_t *chandle = NULL;
snd_hctl_t *handle = NULL; 

snd_ctl_elem_value_t *eval = NULL;
snd_hctl_elem_t *elem = NULL;

int DEBUG=0;


/* Initialise ALSA Stuff*/
int alsa_prepare(char* hw, char* ctl_elem_name) {

    int err=0;
    int j=0;
    int ctls_num;

    snd_ctl_elem_id_t *e_id=NULL;
       snd_ctl_elem_info_t *einfo;
    snd_ctl_card_info_t *info;
    snd_ctl_elem_list_t *clist;
    snd_hctl_elem_t *ielem = NULL;

    if ((err = snd_ctl_open(&chandle, hw, 0)) < 0) {
            fprintf(stderr,"ALSA open failed: %s\n",snd_strerror(err));
            return -1;
    }
    
    if ((err = snd_hctl_open(&handle, hw,0)) < 0) {
            fprintf(stderr,"ALSA hctl open failed: %s\n",snd_strerror(err));
            return -2;
    }

    if ((err = snd_ctl_card_info_malloc(&info)) < 0) {
            fprintf(stderr, "ALSA ctl card info malloc failed: %s\n",
                            snd_strerror(err));
            return -3;
    }
    memset(info,0,snd_ctl_card_info_sizeof());

    if ((err = snd_ctl_card_info(chandle, info)) < 0) {
            fprintf(stderr,"ALSA hw info failed: %s\n",snd_strerror(err));
            return -4;
    }

    if ((err = snd_ctl_elem_list_malloc(&clist)) < 0) {
            fprintf(stderr,"ALSA ctl elem list malloc failed: %s\n",
                            snd_strerror(err));
            return -5;
    }
    memset(clist,0,snd_ctl_elem_list_sizeof());

    if ((err = snd_ctl_elem_list(chandle,clist)) < 0) {
            fprintf(stderr,"ALSA ctl_elem list get failed: %s\n",
                            snd_strerror(err));
            return -6;
    }

    if (DEBUG > 0) {
        printf("Card %s (%s)\n",
            snd_ctl_card_info_get_id(info),
            snd_ctl_card_info_get_longname(info));
    }
    
    ctls_num=snd_ctl_elem_list_get_count(clist);
    snd_ctl_elem_list_free(clist);

    if ((err = snd_hctl_load(handle)) < 0) {
            fprintf(stderr,"ALSA hctl load failed: %s\n",snd_strerror(err));
            return -7;
    }
    ielem = snd_hctl_first_elem(handle);

    for( j=0 ; j<ctls_num ; j++ ) {
        if ((err = snd_ctl_elem_id_malloc(&e_id)) < 0) {
                fprintf(stderr,"ALSA ctl elem id malloc failed: %s\n",
                                snd_strerror(err));
                return -8;
        }
                
        snd_hctl_elem_get_id(ielem,e_id);

        if (DEBUG > 0) {
            printf("#%d '%s'%d\n",
                snd_ctl_elem_id_get_numid(e_id),
                snd_ctl_elem_id_get_name(e_id),
                snd_ctl_elem_id_get_index(e_id));
        }

        if (strcmp(ctl_elem_name, snd_ctl_elem_id_get_name(e_id)) == 0)
        {
            if (DEBUG > 0) {
                printf("Found '%s'\n", ctl_elem_name);
            }

            if ((err = snd_ctl_elem_info_malloc(&einfo)) < 0) {
                fprintf(stderr,"ALSA ctl_elem_info_malloc failed: %s\n",
                                snd_strerror(err));
                return -10;
            }
            memset(einfo,0,snd_ctl_elem_info_sizeof());
            snd_ctl_elem_info_set_id(einfo,e_id);
            if ((err = snd_hctl_elem_info(ielem,einfo)) < 0) {
                fprintf(stderr,"ALSA control info read failed: %s\n",
                                snd_strerror(err));
                return -11;
            }
    
            if ((err = snd_ctl_elem_value_malloc(&eval)) < 0) {
                fprintf(stderr,"ALSA snd_ctl_elem_value_malloc failed: %s\n",
                                snd_strerror(err));
                return -12;
            }

            memset(eval,0,snd_ctl_elem_value_sizeof());
            snd_ctl_elem_value_set_id(eval,e_id);

            elem = ielem;
        }

        ielem=snd_hctl_elem_next(ielem);
    }

    if (elem == NULL) {
            fprintf(stderr, "Haven't found '%s' on card '%s'!\n\n",
                            ctl_elem_name,
                            snd_ctl_card_info_get_longname(info));
            return -100;
    }

    return 0;
}

/* Set Mixer Values */
int update_mixer(int change) {
    int val;

    if (elem == NULL) {
        printf("Error, elem undefined!\n");
        exit(-1);
    }

    if (eval == NULL) {
        printf("Error, eval undefined!\n");
        exit(-2);
    }

    snd_hctl_elem_read(elem,eval);

    val = snd_ctl_elem_value_get_integer(eval, 0);
    if (DEBUG > 0) {
        printf("val0: %d\n", val);
    }
    snd_ctl_elem_value_set_integer(eval, 0, (val+change));
    
    val = snd_ctl_elem_value_get_integer(eval, 1);
    if (DEBUG > 0) {
        printf("val1: %d\n", val);
    }
    snd_ctl_elem_value_set_integer(eval, 1, (val+change));

    snd_ctl_elem_write(chandle, eval);
        
    return 0;
}

/* Close ALSA handles */
int alsa_halt(void) {
    snd_ctl_close(chandle);
    snd_hctl_close(handle);

    return 0;
}

/* Wait for input from event node and update mixer */
int loop(int fd) {
    int i;           /* loop counter */
    size_t read_bytes;  /* how many bytes were read */
    struct input_event ev[64]; /* the events (up to 64 at once) */

    while (1) {
        read_bytes = read(fd, ev, sizeof(struct input_event) * 64);

        if (read_bytes < (int) sizeof(struct input_event)) {
            perror("evtest: short read");
            exit (1);
        }

        for (i = 0; i < (int) (read_bytes / sizeof(struct
                        input_event)); i++)
        {
            if(ev[i].type==EV_KEY &&
                    ev[i].code==KEY_VOLUMEUP) {
                if(ev[i].value==1)
                    /*      printf("UP\n"); */
                    update_mixer(1);
            } else if(ev[i].type==EV_KEY &&
                    ev[i].code==KEY_VOLUMEDOWN) {
                if(ev[i].value==1)
                    /*      printf("DOWN\n"); */
                    update_mixer(-1);
            }
            
            if (DEBUG > 1) {
                printf("Event: time %ld.%06ld, type %d, "
                        "code %d, value %d\n",
                        ev[i].time.tv_sec,
                        ev[i].time.tv_usec, 
                        ev[i].type,
                        ev[i].code,
                        ev[i].value);
            }
        }
    }
}

int print_help(char *progname) {

    fprintf(stderr, "usage: %s [options]\n\n", progname);
    fprintf(stderr, "Possible Options:\n");
    fprintf(stderr, "-d: Debug (more -d, more Debug)\n");
    fprintf(stderr, "-p: Specify PID File, where child PID should "
                "be written to\n\n");
    fprintf(stderr, "-e: Input Event Device:\n");
    fprintf(stderr, "    usually /dev/input/event?\n\n");
    fprintf(stderr, "-s: ALSA Hardware\n");
    fprintf(stderr, "    usually hw:0 for first Soundcard, "
                        "hw:1 for second...\n\n");
    fprintf(stderr, "-n: Alsa Subdevice to use. Should be something like \n");
    fprintf(stderr, "    'Speaker Playback Volume'.\n\n");
    fprintf(stderr, "-v: Print Version\n\n");
    return 0;
}


int main(int argc, char **argv) {

    char *sndhw = NULL;
    char *input_event = NULL;
    char *ctl_elem_name = NULL;
    char *pid_file = NULL;

    pid_t pid;

    int fd = -1;        /* the file descriptor for the device */
    int c=0;
    int err=0;

    while ((c = getopt(argc, argv, "dhve:n:s:p:")) != -1) {
        switch (c) {
            case 'e':
                    input_event = optarg;
                    break;
                    
            case 's':
                    sndhw = optarg;
                    break;

            case 'n':
                    ctl_elem_name = optarg;
                    break;
                    
            case 'd':
                DEBUG++;
                break;

            case 'h':
                print_help(argv[0]);
                exit(0);
                break;

        case 'v':
        fprintf(stderr, "%s: Version %s\n\n", argv[0], VERSION);
        exit(0);
        break;

        case 'p':
        pid_file = optarg;
        break;

            default:
                print_help(argv[0]);
                exit(-2);
        }

    }

    if (input_event == NULL) 
            input_event = strdup("/dev/input/event0");

    if (sndhw == NULL)
            sndhw = strdup("hw:0");

    if (ctl_elem_name == NULL)
            ctl_elem_name = strdup("Speaker Playback Volume");

    if ((fd = open(input_event, O_RDONLY)) < 0) {
        fprintf(stderr, "Can't open %s\n", input_event);
        exit(-3);
    }

    /* If not debug, fork into background */
    if (DEBUG == 0) {
        switch (pid = fork())         /* fork the process */
        {
            case -1:            /* error */
                fprintf(stderr, "Fork failed!\n");
                exit(-80);
                break;

            case 0:
                /* Child, continue execution */
                            break;

            default:
                /* Father process, check for fork, then exit */
                if (pid_file != NULL) {
                    FILE *ppidfile;
                    if ((ppidfile = fopen(pid_file, "w")) != NULL) {
                        fprintf(ppidfile, "%d\n", pid);
			fclose(ppidfile);
		    }
		}
                exit(0);
                break;
        }
    }
    
    if ((err = alsa_prepare(sndhw, ctl_elem_name))) {
            exit(err);
    }

    loop(fd);

    close(fd);

    alsa_halt();
    return 0;
}
